@@ -26,9 +26,8 @@ import qualified Data.Text as T
 import qualified Text.Pandoc.Builder as B
 import Text.Pandoc.Class.PandocMonad (PandocMonad (..))
 import Text.Pandoc.Definition
-import Text.Pandoc.Error (PandocError (PandocParsecError))
 import Text.Pandoc.Options
-import Text.Pandoc.Parsing hiding (enclosed, nested)
+import Text.Pandoc.Parsing hiding (enclosed)
 import Text.Pandoc.Shared (trim, stringify, tshow)
 import Data.List (isPrefixOf, isSuffixOf)
 import qualified Safe
@@ -43,26 +42,17 @@ readDokuWiki opts s = do
   res <- runParserT parseDokuWiki def {stateOptions = opts }
            (initialSourceName sources) sources
   case res of
-       Left e  -> throwError $ PandocParsecError sources e
+       Left e  -> throwError $ fromParsecError sources e
        Right d -> return d
 
-type DWParser = ParserT Sources ParserState
+type DWParser = ParsecT Sources ParserState
 
 -- * Utility functions
 
 -- | Parse end-of-line, which can be either a newline or end-of-file.
-eol :: (Stream s m Char, UpdateSourcePos s Char) => ParserT s st m ()
+eol :: (Stream s m Char, UpdateSourcePos s Char) => ParsecT s st m ()
 eol = void newline <|> eof
 
-nested :: PandocMonad m => DWParser m a -> DWParser m a
-nested p = do
-  nestlevel <- stateMaxNestingLevel <$>  getState
-  guard $ nestlevel > 0
-  updateState $ \st -> st{ stateMaxNestingLevel = stateMaxNestingLevel st - 1 }
-  res <- p
-  updateState $ \st -> st{ stateMaxNestingLevel = nestlevel }
-  return res
-
 guardColumnOne :: PandocMonad m => DWParser m ()
 guardColumnOne = getPosition >>= \pos -> guard (sourceColumn pos == 1)
 
@@ -117,8 +107,8 @@ inline'' = br
       <|> footnote
       <|> inlineCode
       <|> inlineFile
-      <|> inlineHtml
-      <|> inlinePhp
+      <|> inlineRaw
+      <|> math
       <|> autoLink
       <|> autoEmail
       <|> notoc
@@ -164,7 +154,7 @@ nestedInlines :: (Show a, PandocMonad m)
 nestedInlines end = innerSpace <|> nestedInline
   where
     innerSpace   = try $ whitespace <* notFollowedBy end
-    nestedInline = notFollowedBy whitespace >> nested inline
+    nestedInline = notFollowedBy whitespace >> inline
 
 bold :: PandocMonad m => DWParser m B.Inlines
 bold = try $ B.strong <$> enclosed (string "**") nestedInlines
@@ -209,11 +199,22 @@ inlineCode = codeTag B.codeWith "code"
 inlineFile :: PandocMonad m => DWParser m B.Inlines
 inlineFile = codeTag B.codeWith "file"
 
-inlineHtml :: PandocMonad m => DWParser m B.Inlines
-inlineHtml = try $ B.rawInline "html" <$ string "<html>" <*> manyTillChar anyChar (try $ string "</html>")
+inlineRaw :: PandocMonad m => DWParser m B.Inlines
+inlineRaw = try $ do
+  char '<'
+  fmt <- oneOfStrings ["html", "php", "latex"]
+  -- LaTeX via https://www.dokuwiki.org/plugin:latex
+  char '>'
+  contents <- manyTillChar anyChar
+                (try $ string "</" *> string (T.unpack fmt) *> char '>')
+  return $
+    case T.toLower fmt of
+         "php" -> B.rawInline "html" $ "<?php " <> contents <> " ?>"
+         f -> B.rawInline f contents
 
-inlinePhp :: PandocMonad m => DWParser m B.Inlines
-inlinePhp = try $ B.codeWith ("", ["php"], []) <$ string "<php>" <*> manyTillChar anyChar (try $ string "</php>")
+-- see https://www.dokuwiki.org/plugin:latex
+math :: PandocMonad m => DWParser m B.Inlines
+math = (B.displayMath <$> mathDisplay) <|> (B.math <$> mathInline)
 
 makeLink :: (Text, Text) -> B.Inlines
 makeLink (text, url) = B.link url "" $ B.str text
@@ -243,7 +244,7 @@ nocache :: PandocMonad m => DWParser m B.Inlines
 nocache = try $ mempty <$ string "~~NOCACHE~~"
 
 str :: PandocMonad m => DWParser m B.Inlines
-str = B.str <$> (many1Char alphaNum <|> countChar 1 characterReference)
+str = B.str <$> (many1Char alphaNum <|> characterReference)
 
 symbol :: PandocMonad m => DWParser m B.Inlines
 symbol = B.str <$> countChar 1 nonspaceChar
@@ -388,7 +389,11 @@ image = try $ parseLink fromRaw "{{" "}}"
         parameterList = T.splitOn "&" $ T.drop 1 parameters
         linkOnly = "linkonly" `elem` parameterList
         (width, height) = maybe (Nothing, Nothing) parseWidthHeight (F.find isWidthHeightParameter parameterList)
-        attributes = catMaybes [fmap ("width",) width, fmap ("height",) height]
+        attributes = catMaybes [
+                fmap ("width",) width,
+                fmap ("height",) height,
+                fmap ("query",) (if T.null parameters then Nothing else Just parameters)
+            ]
         defaultDescription = B.str $ urlToText path'
 
 -- * Block parsers
@@ -410,8 +415,7 @@ blockElements = horizontalLine
             <|> quote
             <|> blockCode
             <|> blockFile
-            <|> blockHtml
-            <|> blockPhp
+            <|> blockRaw
             <|> table
 
 horizontalLine :: PandocMonad m => DWParser m B.Blocks
@@ -462,17 +466,19 @@ quote = try $ nestedQuote 0
     quoteContinuation level = mconcat <$> many (try $ prefix level *> contents level)
     nestedQuote level = B.blockQuote <$ char '>' <*> quoteContents (level + 1 :: Int)
 
-blockHtml :: PandocMonad m => DWParser m B.Blocks
-blockHtml = try $ B.rawBlock "html"
-  <$  string "<HTML>"
-  <*  optional (manyTill spaceChar eol)
-  <*> manyTillChar anyChar (try $ string "</HTML>")
-
-blockPhp :: PandocMonad m => DWParser m B.Blocks
-blockPhp = try $ B.codeBlockWith ("", ["php"], [])
-  <$  string "<PHP>"
-  <*  optional (manyTill spaceChar eol)
-  <*> manyTillChar anyChar (try $ string "</PHP>")
+blockRaw :: PandocMonad m => DWParser m B.Blocks
+blockRaw = try $ do
+  char '<'
+  fmt <- oneOfStrings ["HTML", "PHP", "LATEX"]
+  -- LaTeX via https://www.dokuwiki.org/plugin:latex
+  char '>'
+  optional (manyTill spaceChar eol)
+  contents <- manyTillChar anyChar
+               (try $ string "</" *> string (T.unpack fmt) *> char '>')
+  return $
+    case T.toLower fmt of
+         "php" -> B.rawBlock "html" $ "<?php " <> contents <> " ?>"
+         f -> B.rawBlock f contents
 
 table :: PandocMonad m => DWParser m B.Blocks
 table = do
