@@ -16,7 +16,7 @@ module Text.Pandoc.Readers.Man (readMan) where
 
 import Data.Char (toLower)
 import Data.Default (Default)
-import Control.Monad (liftM, mzero, guard, void)
+import Control.Monad (mzero, guard, void)
 import Control.Monad.Trans (lift)
 import Control.Monad.Except (throwError)
 import Data.Maybe (catMaybes, isJust)
@@ -24,15 +24,12 @@ import Data.List (intersperse)
 import qualified Data.Text as T
 import Text.Pandoc.Builder as B
 import Text.Pandoc.Class.PandocMonad (PandocMonad(..), report)
-import Text.Pandoc.Error (PandocError (PandocParsecError))
 import Text.Pandoc.Logging (LogMessage(..))
 import Text.Pandoc.Options
 import Text.Pandoc.Parsing
 import Text.Pandoc.Walk (query)
-import Text.Pandoc.Shared (mapLeft)
 import Text.Pandoc.Readers.Roff  -- TODO explicit imports
-import qualified Text.Parsec as Parsec
-import Text.Parsec.Pos (updatePosString)
+import qualified Text.Pandoc.Parsing as P
 import qualified Data.Foldable as Foldable
 
 data ManState = ManState { readerOptions   :: ReaderOptions
@@ -45,7 +42,7 @@ instance Default ManState where
                  , metadata        = nullMeta
                  , tableCellsPlain = True }
 
-type ManParser m = ParserT [RoffToken] ManState m
+type ManParser m = P.ParsecT [RoffToken] ManState m
 
 
 -- | Read man (troff) from an input string and return a Pandoc document.
@@ -57,21 +54,18 @@ readMan opts s = do
   let Sources inps = toSources s
   tokenz <- mconcat <$> mapM (uncurry lexRoff) inps
   let state = def {readerOptions = opts} :: ManState
-  let fixError (PandocParsecError _ e) = PandocParsecError (Sources inps) e
-      fixError e = e
   eitherdoc <- readWithMTokens parseMan state
      (Foldable.toList . unRoffTokens $ tokenz)
-  either (throwError . fixError) return eitherdoc
+  either (throwError . fromParsecError (Sources inps)) return eitherdoc
 
 
 readWithMTokens :: PandocMonad m
-        => ParserT [RoffToken] ManState m a  -- ^ parser
+        => ParsecT [RoffToken] ManState m a  -- ^ parser
         -> ManState                         -- ^ initial state
         -> [RoffToken]                       -- ^ input
-        -> m (Either PandocError a)
+        -> m (Either ParseError a)
 readWithMTokens parser state input =
-  let leftF = PandocParsecError mempty
-  in mapLeft leftF `liftM` runParserT parser state "source" input
+  runParserT parser state "source" input
 
 
 parseMan :: PandocMonad m => ManParser m Pandoc
@@ -180,14 +174,16 @@ parseNewParagraph = do
 -- Parser: [RoffToken] -> Pandoc
 --
 
-msatisfy :: Monad m => (RoffToken -> Bool) -> ParserT [RoffToken] st m RoffToken
-msatisfy predic = tokenPrim show nextPos testTok
+msatisfy :: Monad m
+         => (RoffToken -> Bool) -> P.ParsecT [RoffToken] st m RoffToken
+msatisfy predic = P.tokenPrim show nextPos testTok
   where
     testTok t     = if predic t then Just t else Nothing
     nextPos _pos _x (ControlLine _ _ pos':_) = pos'
-    nextPos pos _x _xs  = updatePosString
-                             (setSourceColumn
-                               (setSourceLine pos $ sourceLine pos + 1) 1) ""
+    nextPos pos _x _xs  = P.updatePosString
+                             (P.setSourceColumn
+                               (P.setSourceLine pos $
+                                 P.sourceLine pos + 1) 1) ""
 
 mtoken :: PandocMonad m => ManParser m RoffToken
 mtoken = msatisfy (const True)
@@ -431,7 +427,7 @@ listItem mbListType = try $ do
     (arg1 : _)  -> do
       let cs = linePartsToText arg1
       let cs' = if not (T.any (== '.') cs || T.any (== ')') cs) then cs <> "." else cs
-      let lt = case Parsec.runParser anyOrderedListMarker defaultParserState
+      let lt = case P.runParser anyOrderedListMarker defaultParserState
                      "list marker" cs' of
                   Right (start, listtype, listdelim)
                     | cs == cs' -> Ordered (start, listtype, listdelim)
