@@ -1,7 +1,7 @@
 {-# LANGUAGE OverloadedStrings #-}
 {- |
    Module      : Text.Pandoc.Readers.MediaWiki
-   Copyright   : Copyright (C) 2012-2022 John MacFarlane
+   Copyright   : Copyright (C) 2012-2023 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -19,7 +19,7 @@ module Text.Pandoc.Readers.MediaWiki ( readMediaWiki ) where
 
 import Control.Monad
 import Control.Monad.Except (throwError)
-import Data.Char (isDigit, isSpace)
+import Data.Char (isDigit, isLetter, isSpace)
 import qualified Data.Foldable as F
 import Data.List (intersperse)
 import Data.Maybe (fromMaybe, maybeToList)
@@ -34,11 +34,10 @@ import Text.Pandoc.Class.PandocMonad (PandocMonad (..))
 import Text.Pandoc.Definition
 import Text.Pandoc.Logging
 import Text.Pandoc.Options
-import Text.Pandoc.Parsing hiding (nested, tableCaption)
-import Text.Pandoc.Readers.HTML (htmlTag, isBlockTag, isCommentTag)
+import Text.Pandoc.Parsing hiding (tableCaption)
+import Text.Pandoc.Readers.HTML (htmlTag, isBlockTag, isCommentTag, toAttr)
 import Text.Pandoc.Shared (safeRead, stringify, stripTrailingNewlines,
-                           trim, splitTextBy, tshow)
-import Text.Pandoc.Walk (walk)
+                           trim, splitTextBy, tshow, formatCode)
 import Text.Pandoc.XML (fromEntities)
 
 -- | Read mediawiki from an input string and return a Pandoc document.
@@ -70,7 +69,7 @@ data MWState = MWState { mwOptions         :: ReaderOptions
                        , mwInTT            :: Bool
                        }
 
-type MWParser m = ParserT Sources MWState m
+type MWParser m = ParsecT Sources MWState m
 
 instance HasReaderOptions MWState where
   extractReaderOptions = mwOptions
@@ -87,17 +86,6 @@ instance HasLogMessages MWState where
 -- auxiliary functions
 --
 
--- This is used to prevent exponential blowups for things like:
--- ''a'''a''a'''a''a'''a''a'''a
-nested :: PandocMonad m => MWParser m a -> MWParser m a
-nested p = do
-  nestlevel <- mwMaxNestingLevel `fmap` getState
-  guard $ nestlevel > 0
-  updateState $ \st -> st{ mwMaxNestingLevel = mwMaxNestingLevel st - 1 }
-  res <- p
-  updateState $ \st -> st{ mwMaxNestingLevel = nestlevel }
-  return res
-
 specialChars :: [Char]
 specialChars = "'[]<=&*{}|\":\\"
 
@@ -223,22 +211,21 @@ table = do
   optional rowsep
   hasheader <- option False $ True <$ lookAhead (skipSpaces *> char '!')
   (cellspecs',hdr) <- unzip <$> tableRow
-  let widths = map ((tableWidth *) . snd) cellspecs'
+  let widths = map (tableWidth *) cellspecs'
   let restwidth = tableWidth - sum widths
   let zerocols = length $ filter (==0.0) widths
   let defaultwidth = if zerocols == 0 || zerocols == length widths
                         then ColWidthDefault
                         else ColWidth $ restwidth / fromIntegral zerocols
   let widths' = map (\w -> if w > 0 then ColWidth w else defaultwidth) widths
-  let cellspecs = zip (map fst cellspecs') widths'
+  let cellspecs = zip (calculateAlignments hdr) widths'
   rows' <- many $ try $ rowsep *> (map snd <$> tableRow)
   optional blanklines
   tableEnd
-  let cols = length hdr
   let (headers,rows) = if hasheader
                           then (hdr, rows')
-                          else (replicate cols mempty, hdr:rows')
-  let toRow = Row nullAttr . map B.simpleCell
+                          else ([], hdr:rows')
+  let toRow = Row nullAttr
       toHeaderRow l = [toRow l | not (null l)]
   return $ B.table (B.simpleCaption $ B.plain caption)
                    cellspecs
@@ -246,6 +233,12 @@ table = do
                    [TableBody nullAttr 0 [] $ map toRow rows]
                    (TableFoot nullAttr [])
 
+calculateAlignments :: [Cell] -> [Alignment]
+calculateAlignments = map cellAligns
+  where
+    cellAligns :: Cell -> Alignment
+    cellAligns (Cell _ align _ _ _) = align
+
 parseAttrs :: PandocMonad m => MWParser m [(Text,Text)]
 parseAttrs = many1 parseAttr
 
@@ -253,7 +246,9 @@ parseAttr :: PandocMonad m => MWParser m (Text, Text)
 parseAttr = try $ do
   skipMany spaceChar
   k <- many1Char letter
+  skipMany spaceChar
   char '='
+  skipMany spaceChar
   v <- (char '"' >> many1TillChar (satisfy (/='\n')) (char '"'))
        <|> many1Char (satisfy $ \c -> not (isSpace c) && c /= '|')
   return (k,v)
@@ -266,7 +261,10 @@ tableEnd = try $ guardColumnOne *> skipSpaces *> sym "|}"
 
 rowsep :: PandocMonad m => MWParser m ()
 rowsep = try $ guardColumnOne *> skipSpaces *> sym "|-" <*
-               many (char '-') <* optional parseAttrs <* blanklines
+               many (char '-') <* optional parseAttrs
+                               <* skipSpaces
+                               <* skipMany htmlComment
+                               <* blanklines
 
 cellsep :: PandocMonad m => MWParser m ()
 cellsep = try $ do
@@ -287,6 +285,7 @@ cellsep = try $ do
 
 tableCaption :: PandocMonad m => MWParser m Inlines
 tableCaption = try $ do
+  optional rowsep
   guardColumnOne
   skipSpaces
   sym "|+"
@@ -294,14 +293,14 @@ tableCaption = try $ do
   trimInlines . mconcat <$>
     many (notFollowedBy (cellsep <|> rowsep) *> inline)
 
-tableRow :: PandocMonad m => MWParser m [((Alignment, Double), Blocks)]
+tableRow :: PandocMonad m => MWParser m [(Double, Cell)]
 tableRow = try $ skipMany htmlComment *> many tableCell
 
-tableCell :: PandocMonad m => MWParser m ((Alignment, Double), Blocks)
+tableCell :: PandocMonad m => MWParser m (Double, Cell)
 tableCell = try $ do
   cellsep
   skipMany spaceChar
-  attrs <- option [] $ try $ parseAttrs <* skipSpaces <* char '|' <*
+  attribs <- option [] $ try $ parseAttrs <* skipSpaces <* char '|' <*
                                  notFollowedBy (char '|')
   skipMany spaceChar
   pos' <- getPosition
@@ -309,15 +308,23 @@ tableCell = try $ do
                             ((snd <$> withRaw table) <|> countChar 1 anyChar))
   bs <- parseFromString (do setPosition pos'
                             mconcat <$> many block) ls
-  let align = case lookup "align" attrs of
+  let align = case lookup "align" attribs of
                     Just "left"   -> AlignLeft
                     Just "right"  -> AlignRight
                     Just "center" -> AlignCenter
                     _             -> AlignDefault
-  let width = case lookup "width" attrs of
+  let width = case lookup "width" attribs of
                     Just xs -> fromMaybe 0.0 $ parseWidth xs
                     Nothing -> 0.0
-  return ((align, width), bs)
+  let rowspan = RowSpan . fromMaybe 1 $
+                safeRead =<< lookup "rowspan" attribs
+  let colspan = ColSpan . fromMaybe 1 $
+                safeRead =<< lookup "colspan" attribs
+  let handledAttribs = ["align", "colspan", "rowspan"]
+      attribs' = [ (k, v) | (k, v) <- attribs
+                          , k `notElem` handledAttribs
+                 ]
+  return (width, B.cellWith (toAttr attribs') align rowspan colspan bs)
 
 parseWidth :: Text -> Maybe Double
 parseWidth s =
@@ -392,14 +399,7 @@ preformatted = try $ do
      else return $ B.para $ encode contents
 
 encode :: Inlines -> Inlines
-encode = B.fromList . normalizeCode . B.toList . walk strToCode
-  where strToCode (Str s) = Code ("",[],[]) s
-        strToCode Space   = Code ("",[],[]) " "
-        strToCode  x      = x
-        normalizeCode []  = []
-        normalizeCode (Code a1 x : Code a2 y : zs) | a1 == a2 =
-          normalizeCode $ Code a1 (x <> y) : zs
-        normalizeCode (x:xs) = x : normalizeCode xs
+encode = formatCode nullAttr
 
 header :: PandocMonad m => MWParser m Blocks
 header = try $ do
@@ -664,7 +664,8 @@ internalLink = try $ do
              -- [[Help:Contents|] -> "Contents"
              <|> return (B.text $ T.drop 1 $ T.dropWhile (/=':') pagename) )
   sym "]]"
-  linktrail <- B.text <$> manyChar letter
+  -- see #8525:
+  linktrail <- B.text <$> manyChar (satisfy (\c -> isLetter c && not (isCJK c)))
   let link = B.link (addUnderscores pagename) "wikilink" (label <> linktrail)
   if "Category:" `T.isPrefixOf` pagename
      then do
@@ -672,6 +673,23 @@ internalLink = try $ do
        return mempty
      else return link
 
+isCJK :: Char -> Bool
+isCJK c =
+  (c >= '\x3400' && c <= '\x4DBF') ||
+  (c >= '\x4E00' && c <= '\x9FFF') ||
+  (c >= '\x20000' && c <= '\x2A6DF') ||
+  (c >= '\x2A700' && c <= '\x2B73F') ||
+  (c >= '\x2B740' && c <= '\x2B81F') ||
+  (c >= '\x2B820' && c <= '\x2CEAF') ||
+  (c >= '\x2CEB0' && c <= '\x2EBEF') ||
+  (c >= '\x30000' && c <= '\x3134F') ||
+  (c >= '\x31350' && c <= '\x323AF') ||
+  (c >= '\xF900' && c <= '\xFAFF') ||
+  (c >= '\x2F800' && c <= '\x2FA1F') ||
+  (c >= '\x2F00' && c <= '\x2FDF') ||
+  (c >= '\x2E80' && c <= '\x2EFF') ||
+  (c >= '\x3000' && c <= '\x303F')
+
 externalLink :: PandocMonad m => MWParser m Inlines
 externalLink = try $ do
   char '['
@@ -695,12 +713,12 @@ inlinesBetween start end =
   trimInlines . mconcat <$> try (start >> many1Till inline end)
 
 emph :: PandocMonad m => MWParser m Inlines
-emph = B.emph <$> nested (inlinesBetween start end)
+emph = B.emph <$> inlinesBetween start end
     where start = sym "''"
           end   = try $ notFollowedBy' (() <$ strong) >> sym "''"
 
 strong :: PandocMonad m => MWParser m Inlines
-strong = B.strong <$> nested (inlinesBetween start end)
+strong = B.strong <$> inlinesBetween start end
     where start = sym "'''"
           end   = sym "'''"
 
@@ -709,6 +727,6 @@ doubleQuotes = do
   guardEnabled Ext_smart
   inTT <- mwInTT <$> getState
   guard (not inTT)
-  B.doubleQuoted <$> nested (inlinesBetween openDoubleQuote closeDoubleQuote)
+  B.doubleQuoted <$> inlinesBetween openDoubleQuote closeDoubleQuote
     where openDoubleQuote = sym "\"" >> lookAhead nonspaceChar
           closeDoubleQuote = try $ sym "\""
