@@ -25,9 +25,12 @@ However, @\<literal style="html">@ tag is used for HTML raw blocks
 even though it is supported only in Emacs Muse.
 -}
 module Text.Pandoc.Writers.Muse (writeMuse) where
+import Control.Monad (zipWithM)
 import Control.Monad.Except (throwError)
 import Control.Monad.Reader
+    ( asks, MonadReader(local), ReaderT(runReaderT) )
 import Control.Monad.State.Strict
+    ( StateT, gets, modify, evalStateT )
 import Data.Char (isAlphaNum, isAsciiLower, isAsciiUpper, isDigit, isSpace)
 import Data.Default
 import Data.List (intersperse, transpose)
@@ -43,6 +46,7 @@ import Text.Pandoc.ImageSize
 import Text.Pandoc.Options
 import Text.DocLayout
 import Text.Pandoc.Shared
+import Text.Pandoc.URI
 import Text.Pandoc.Templates (renderTemplate)
 import Text.Pandoc.Writers.Math
 import Text.Pandoc.Writers.Shared
@@ -275,7 +279,8 @@ blockToMuse (Table _ blkCapt specs thead tbody tfoot) =
               (length aligns :| length widths : map length (headers:rows))
     isSimple = onlySimpleTableCells (headers : rows) && all (== 0) widths
 blockToMuse (Div _ bs) = flatBlockListToMuse bs
-blockToMuse Null = return empty
+blockToMuse (Figure attr capt body) = do
+  blockToMuse (figureDiv attr capt body)
 
 -- | Return Muse representation of notes collected so far.
 currentNotesToMuse :: PandocMonad m
