@@ -1,4 +1,3 @@
-{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE OverloadedStrings #-}
 {- |
    Module      : Tests.Readers.JATS
@@ -13,7 +12,6 @@ Tests for the JATS reader.
 -}
 module Tests.Readers.JATS (tests) where
 
-import Prelude
 import Data.Text (Text)
 import Test.Tasty
 import Tests.Helpers
@@ -21,13 +19,14 @@ import Text.Pandoc
 import Text.Pandoc.Arbitrary ()
 import Text.Pandoc.Builder
 
+import qualified Data.Text as T
+
 jats :: Text -> Pandoc
 jats = purely $ readJATS def
 
 tests :: [TestTree]
 tests = [ testGroup "inline code"
           [ test jats "basic" $ "<p>\n  <monospace>@&amp;</monospace>\n</p>" =?> para (code "@&")
-          , test jats "lang" $ "<p>\n  <code language=\"c\">@&amp;</code>\n</p>" =?> para (codeWith ("", ["c"], []) "@&")
           ]
         , testGroup "block code"
           [ test jats "basic" $ "<preformat>@&amp;</preformat>" =?> codeBlock "@&"
@@ -88,6 +87,7 @@ tests = [ testGroup "inline code"
             "<p>\n\
             \  <inline-formula><alternatives>\n\
             \  <tex-math><![CDATA[\\sigma|_{\\{x\\}}]]></tex-math>\n\
+            \  </alternatives></inline-formula>\n\
             \</p>"
             =?> para (math "\\sigma|_{\\{x\\}}")
           , test jats "math ml only" $
@@ -126,4 +126,19 @@ tests = [ testGroup "inline code"
             \</sec>"
             =?> header 1 (image "imgs/foo.jpg" "" mempty)
           ]
+
+        , testGroup "metadata"
+          [ test jats "abstract" $
+            T.unlines [ "<front>"
+                      , "<article-meta>"
+                      , "<abstract>"
+                      , "<p>Paragraph 1</p>"
+                      , "<p>Paragraph 2</p>"
+                      , "</abstract>"
+                      , "</article-meta>"
+                      , "</front>"
+                      ] =?>
+            let abstract = para "Paragraph 1" <> para "Paragraph 2"
+            in setMeta "abstract" abstract $ doc mempty
+          ]
         ]
